BIOGEME_OBJECT.PARAMETERS['BTRArmijoBeta1'] = "0.1"
BIOGEME_OBJECT.PARAMETERS['BTRArmijoBeta2'] = "0.9"
BIOGEME_OBJECT.PARAMETERS['BTREta1'] = "0.01"
BIOGEME_OBJECT.PARAMETERS['BTREta2'] = "0.9"
BIOGEME_OBJECT.PARAMETERS['BTRGamma1'] = "0.5"
BIOGEME_OBJECT.PARAMETERS['BTRGamma2'] = "0.5"
BIOGEME_OBJECT.PARAMETERS['BTRIncreaseTRRadius'] = "2"
BIOGEME_OBJECT.PARAMETERS['BTRInitRadius'] = "1"
BIOGEME_OBJECT.PARAMETERS['BTRKappaEpp'] = "0.25"
BIOGEME_OBJECT.PARAMETERS['BTRKappaFrd'] = "0.5"
BIOGEME_OBJECT.PARAMETERS['BTRKappaLbs'] = "0.9"
BIOGEME_OBJECT.PARAMETERS['BTRKappaUbs'] = "0.1"
BIOGEME_OBJECT.PARAMETERS['BTRMaxTRRadius'] = "1e+10"
BIOGEME_OBJECT.PARAMETERS['BTRMinTRRadius'] = "1e-07"
BIOGEME_OBJECT.PARAMETERS['BTRTolerance'] = "6.05545e-06"
BIOGEME_OBJECT.PARAMETERS['BTRTypf'] = "1"
BIOGEME_OBJECT.PARAMETERS['MetropolisHastingsNeighborhoodSize'] = "1"
BIOGEME_OBJECT.PARAMETERS['NormalTruncation'] = "1.96"
BIOGEME_OBJECT.PARAMETERS['TSExpTheta'] = "0.5"
BIOGEME_OBJECT.PARAMETERS['TSFractionGradientRequired'] = "0.1"
BIOGEME_OBJECT.PARAMETERS['TolSchnabelEskow'] = "0.00492157"
BIOGEME_OBJECT.PARAMETERS['cfsqpEps'] = "6.05545e-06"
BIOGEME_OBJECT.PARAMETERS['cfsqpEpsEqn'] = "6.05545e-06"
BIOGEME_OBJECT.PARAMETERS['cfsqpUdelta'] = "0"
BIOGEME_OBJECT.PARAMETERS['missingValue'] = "99999"
BIOGEME_OBJECT.PARAMETERS['sensitivityAnalysisAlpha'] = "0.05"
BIOGEME_OBJECT.PARAMETERS['singularValueThreshold'] = "0.0001"
BIOGEME_OBJECT.PARAMETERS['solvoptErrorArgument'] = "0.0001"
BIOGEME_OBJECT.PARAMETERS['solvoptErrorFunction'] = "1e-06"
BIOGEME_OBJECT.PARAMETERS['tTestThreshold'] = "1.96"
BIOGEME_OBJECT.PARAMETERS['toleranceCheckDerivatives'] = "0.001"
BIOGEME_OBJECT.PARAMETERS['HeaderForWeightInSimulatedResults'] = "__weight"
BIOGEME_OBJECT.PARAMETERS['HeaderOfRowId'] = "__rowId__"
BIOGEME_OBJECT.PARAMETERS['OutputFileForSensitivityAnalysis'] = "__sensitivity.dat"
BIOGEME_OBJECT.PARAMETERS['RandomDistribution'] = "MLHS"
BIOGEME_OBJECT.PARAMETERS['individualId'] = "__rowId__"
BIOGEME_OBJECT.PARAMETERS['optimizationAlgorithm'] = "BIO"
BIOGEME_OBJECT.PARAMETERS['stopFileName'] = "STOP"
BIOGEME_OBJECT.PARAMETERS['warningSign'] = "*"
BIOGEME_OBJECT.PARAMETERS['BTRCheapHessian'] = "1"
BIOGEME_OBJECT.PARAMETERS['BTRExactHessian'] = "1"
BIOGEME_OBJECT.PARAMETERS['BTRIncreaseDraws'] = "2"
BIOGEME_OBJECT.PARAMETERS['BTRInitQuasiNewtonWithBHHH'] = "1"
BIOGEME_OBJECT.PARAMETERS['BTRInitQuasiNewtonWithTrueHessian'] = "1"
BIOGEME_OBJECT.PARAMETERS['BTRMaxIter'] = "1000"
BIOGEME_OBJECT.PARAMETERS['BTRQuasiNewtonUpdate'] = "1"
BIOGEME_OBJECT.PARAMETERS['BTRSignificantDigits'] = "7"
BIOGEME_OBJECT.PARAMETERS['BTRStartDraws'] = "10"
BIOGEME_OBJECT.PARAMETERS['BTRUnfeasibleCGIterations'] = "0"
BIOGEME_OBJECT.PARAMETERS['BTRUsePreconditioner'] = "1"
BIOGEME_OBJECT.PARAMETERS['DumpSensitivityAnalysis'] = "0"
BIOGEME_OBJECT.PARAMETERS['NbrOfDraws'] = "500"
BIOGEME_OBJECT.PARAMETERS['NbrOfDrawsForSensitivityAnalysis'] = "100"
BIOGEME_OBJECT.PARAMETERS['Seed'] = "9021967"
BIOGEME_OBJECT.PARAMETERS['accessFirstDataFromMetaIterator'] = "0"
BIOGEME_OBJECT.PARAMETERS['allowNestedMonteCarlo'] = "0"
BIOGEME_OBJECT.PARAMETERS['biogemeDisplay'] = "0"
BIOGEME_OBJECT.PARAMETERS['biogemeLogDisplay'] = "1"
BIOGEME_OBJECT.PARAMETERS['buildAnalyticalGradient'] = "1"
BIOGEME_OBJECT.PARAMETERS['cfsqpIprint'] = "2"
BIOGEME_OBJECT.PARAMETERS['cfsqpMaxIter'] = "1000"
BIOGEME_OBJECT.PARAMETERS['cfsqpMode'] = "100"
BIOGEME_OBJECT.PARAMETERS['checkDerivatives'] = "0"
BIOGEME_OBJECT.PARAMETERS['computeInitLoglikelihood'] = "1"
BIOGEME_OBJECT.PARAMETERS['decimalPrecisionForSimulation'] = "6"
BIOGEME_OBJECT.PARAMETERS['deriveAnalyticalHessian'] = "1"
BIOGEME_OBJECT.PARAMETERS['dumpDrawsOnFile'] = "0"
BIOGEME_OBJECT.PARAMETERS['firstIdOfLiterals'] = "10000"
BIOGEME_OBJECT.PARAMETERS['includeDataWhenSimulate'] = "0"
BIOGEME_OBJECT.PARAMETERS['maxGcpIter'] = "10000"
BIOGEME_OBJECT.PARAMETERS['maxPrimeNumbers'] = "1000"
BIOGEME_OBJECT.PARAMETERS['monteCarloControlVariateReport'] = "100"
BIOGEME_OBJECT.PARAMETERS['moreRobustToNumericalIssues'] = "0"
BIOGEME_OBJECT.PARAMETERS['numberOfThreads'] = "12"
BIOGEME_OBJECT.PARAMETERS['printPValue'] = "1"
BIOGEME_OBJECT.PARAMETERS['saveUniformDraws'] = "1"
BIOGEME_OBJECT.PARAMETERS['scaleDerivativesInSums'] = "0"
BIOGEME_OBJECT.PARAMETERS['shareOfProcessors'] = "50"
BIOGEME_OBJECT.PARAMETERS['simulateReportForEveryObservation'] = "1"
BIOGEME_OBJECT.PARAMETERS['solvoptDisplay'] = "1"
BIOGEME_OBJECT.PARAMETERS['solvoptMaxIter'] = "1000"
BIOGEME_OBJECT.PARAMETERS['svdMaxIter'] = "150"
BIOGEME_OBJECT.PARAMETERS['useAnalyticalHessianForOptimization'] = "0"
BIOGEME_OBJECT.PARAMETERS['varCovarFromBHHH'] = "0"
BIOGEME_OBJECT.PARAMETERS['warnsForIllegalElements'] = "1"

