from biogeme.expressions import *
GROUP=Variable('GROUP')
SURVEY=Variable('SURVEY')
SP=Variable('SP')
ID=Variable('ID')
PURPOSE=Variable('PURPOSE')
FIRST=Variable('FIRST')
TICKET=Variable('TICKET')
WHO=Variable('WHO')
LUGGAGE=Variable('LUGGAGE')
AGE=Variable('AGE')
MALE=Variable('MALE')
INCOME=Variable('INCOME')
GA=Variable('GA')
ORIGIN=Variable('ORIGIN')
DEST=Variable('DEST')
TRAIN_AV=Variable('TRAIN_AV')
CAR_AV=Variable('CAR_AV')
SM_AV=Variable('SM_AV')
TRAIN_TT=Variable('TRAIN_TT')
TRAIN_CO=Variable('TRAIN_CO')
TRAIN_HE=Variable('TRAIN_HE')
SM_TT=Variable('SM_TT')
SM_CO=Variable('SM_CO')
SM_HE=Variable('SM_HE')
SM_SEATS=Variable('SM_SEATS')
CAR_TT=Variable('CAR_TT')
CAR_CO=Variable('CAR_CO')
CHOICE=Variable('CHOICE')
