"""Defines a generic exception for Biogeme

:author: Michel Bierlaire

:date: Tue Mar 26 16:47:11 2019

"""
class biogemeError(Exception):
    """ Defines a generic exception for Biogeme. """ 
    pass
